# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/scaffolding/resource_mgmt.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'mytorch/scaffolding/resource_mgmt.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"]\n\x0fResourceRequest\x12\x12\n\nauth_token\x18\x01 \x01(\t\x12\x11\n\tclient_ip\x18\x02 \x01(\t\x12\x11\n\tserver_id\x18\x03 \x01(\t\x12\x10\n\x08max_time\x18\x04 \x01(\x05\"\xe4\x01\n\x17ResourceRequestResponse\x12\x14\n\x0c\x61\x63\x63\x65ss_token\x18\x01 \x01(\t\x12\x17\n\x0f\x61\x64\x64itional_info\x18\x02 \x01(\t\x12\x13\n\x0bresource_id\x18\x03 \x01(\t\x12\x13\n\x0bresource_ip\x18\x04 \x01(\t\x12\x15\n\rresource_port\x18\x05 \x01(\x05\x12\x15\n\rresource_name\x18\x06 \x01(\t\x12\x12\n\nlease_time\x18\x07 \x01(\x05\x12\x17\n\x0fmytorch_version\x18\x08 \x01(\t\x12\x15\n\rresource_info\x18\t \x01(\t2e\n\x13ResourceMgmtService\x12N\n\x10resource_request\x12\x18.mytorch.ResourceRequest\x1a .mytorch.ResourceRequestResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.scaffolding.resource_mgmt_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_RESOURCEREQUEST']._serialized_start=76
  _globals['_RESOURCEREQUEST']._serialized_end=169
  _globals['_RESOURCEREQUESTRESPONSE']._serialized_start=172
  _globals['_RESOURCEREQUESTRESPONSE']._serialized_end=400
  _globals['_RESOURCEMGMTSERVICE']._serialized_start=402
  _globals['_RESOURCEMGMTSERVICE']._serialized_end=503
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from connection_utils.server_connection import ServerConnection
from gRPC_impl import shared_msg_types_pb2
from gRPC_impl.mytorch.nn import module_pb2_grpc
from utils.logger import Logger
from torch.Tensor import Tensor  

class ParametersGenerator:
    """
    Wraps a model UUID, and when iterated, fetches parameters from the server.
    BROKEN !!!
    """
    #def __init__(self, uuid: str):
    #    self.uuid = uuid
    #    self.channel = ServerConnection.get_active_connection()
    #    self.stub = module_pb2_grpc.ModuleServiceStub(self.channel)
    #    self.logger = Logger.get_logger()

    #def __iter__(self):
    #    # Get all parameters in a single gRPC call
    #    request = shared_msg_types_pb2.UUID(uuid=self.uuid)
    #    response = self.stub.GetParameters(request)

    #    # The response is a Parameters message with repeated GrpcTensor fields
    #    for grpc_tensor in response.parameters:
    #        yield Tensor(
    #            uuid=grpc_tensor.uuid,
    #            shape=list(grpc_tensor.shape),
    #            dtype=grpc_tensor.dtype
    #        )



# This is the setup.py file for the mytorch_client package.
# It is used to install the package for local development.
# It is not used to install the package for production.
#
# To install the package for local development, run the following command:
# pip install -e .
#

from setuptools import setup, find_packages

setup(
    name="mytorch_client",
    version="0.4.2",
    packages=find_packages(),
    install_requires=[
        "grpcio>=1.71.0",
        "grpcio-tools>=1.71.0",
        "numpy",
        "requests",
        "pillow",
        "tqdm",
        "matplotlib",
        "scikit-learn",
        "setuptools",
        "typing_extensions",
    ]
) 



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/utils/data/data.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dmytorch/utils/data/data.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"2\n\nGrpcSubset\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x16\n\x0e\x64\x61taset_length\x18\x02 \x01(\x05\"J\n\x0eGrpcDataLoader\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x16\n\x0e\x64\x61taset_length\x18\x02 \x01(\x05\x12\x12\n\nbatch_size\x18\x03 \x01(\x05\"\x9e\x01\n\x17\x44\x61taLoaderBatchResponse\x12)\n\rinputs_tensor\x18\x01 \x01(\x0b\x32\x12.shared.GrpcTensor\x12*\n\x0etargets_tensor\x18\x02 \x01(\x0b\x32\x12.shared.GrpcTensor\x12\x15\n\ris_last_batch\x18\x03 \x01(\x08\x12\x15\n\rout_of_bounds\x18\x04 \x01(\x08\"<\n\x13\x43reateSubsetRequest\x12\x14\n\x0c\x64\x61taset_uuid\x18\x01 \x01(\t\x12\x0f\n\x07indices\x18\x02 \x03(\x05\"T\n\x17\x43reateDataLoaderRequest\x12\x14\n\x0c\x64\x61taset_uuid\x18\x01 \x01(\t\x12\x12\n\nbatch_size\x18\x02 \x01(\x05\x12\x0f\n\x07shuffle\x18\x03 \x01(\x08\"C\n\x13GetNextBatchRequest\x12\x17\n\x0f\x64\x61taloader_uuid\x18\x01 \x01(\t\x12\x13\n\x0b\x62\x61tch_index\x18\x02 \x01(\x05\x32\x89\x02\n\x0b\x44\x61taService\x12I\n\x14\x43reateSubsetOnServer\x12\x1c.mytorch.CreateSubsetRequest\x1a\x13.mytorch.GrpcSubset\x12U\n\x18\x43reateDataLoaderOnServer\x12 .mytorch.CreateDataLoaderRequest\x1a\x17.mytorch.GrpcDataLoader\x12X\n\x16GetNextDataLoaderBatch\x12\x1c.mytorch.GetNextBatchRequest\x1a .mytorch.DataLoaderBatchResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.utils.data.data_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GRPCSUBSET']._serialized_start=66
  _globals['_GRPCSUBSET']._serialized_end=116
  _globals['_GRPCDATALOADER']._serialized_start=118
  _globals['_GRPCDATALOADER']._serialized_end=192
  _globals['_DATALOADERBATCHRESPONSE']._serialized_start=195
  _globals['_DATALOADERBATCHRESPONSE']._serialized_end=353
  _globals['_CREATESUBSETREQUEST']._serialized_start=355
  _globals['_CREATESUBSETREQUEST']._serialized_end=415
  _globals['_CREATEDATALOADERREQUEST']._serialized_start=417
  _globals['_CREATEDATALOADERREQUEST']._serialized_end=501
  _globals['_GETNEXTBATCHREQUEST']._serialized_start=503
  _globals['_GETNEXTBATCHREQUEST']._serialized_end=570
  _globals['_DATASERVICE']._serialized_start=573
  _globals['_DATASERVICE']._serialized_end=838
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from proxies.mytorch.hub.hub_proxy import HubProxy
from torch.nn.Module import Module
from torchvision.models import WeightsEnum

# example call: torch.hub.load('pytorch/vision:v0.6.0', 'resnet18', weights=ResNet18_Weights.DEFAULT)
def load(repo: str, model_name: str, weights: WeightsEnum) -> Module:
    model_uuid =  HubProxy().load_module_from_repo(repo, model_name, weights)
    module = Module(uuid=model_uuid)
    return module



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class ModelLoadRequest(_message.Message):
    __slots__ = ("repo", "model_name", "weights_enum")
    REPO_FIELD_NUMBER: _ClassVar[int]
    MODEL_NAME_FIELD_NUMBER: _ClassVar[int]
    WEIGHTS_ENUM_FIELD_NUMBER: _ClassVar[int]
    repo: str
    model_name: str
    weights_enum: str
    def __init__(self, repo: _Optional[str] = ..., model_name: _Optional[str] = ..., weights_enum: _Optional[str] = ...) -> None: ...

class ModelLoadResponse(_message.Message):
    __slots__ = ("model_uuid",)
    MODEL_UUID_FIELD_NUMBER: _ClassVar[int]
    model_uuid: str
    def __init__(self, model_uuid: _Optional[str] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch_neuronx/mytorch_neuronx.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%mytorch_neuronx/mytorch_neuronx.proto\x12\x0fmytorch_neuronx\"=\n\x0cTraceRequest\x12\x12\n\nmodel_uuid\x18\x01 \x01(\t\x12\x19\n\x11tensor_uuid_tuple\x18\x02 \x03(\t\"#\n\rTraceResponse\x12\x12\n\nmodel_uuid\x18\x01 \x01(\t2a\n\x15\x43qtorchNeuronxService\x12H\n\x05trace\x12\x1d.mytorch_neuronx.TraceRequest\x1a\x1e.mytorch_neuronx.TraceResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch_neuronx.mytorch_neuronx_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TRACEREQUEST']._serialized_start=58
  _globals['_TRACEREQUEST']._serialized_end=119
  _globals['_TRACERESPONSE']._serialized_start=121
  _globals['_TRACERESPONSE']._serialized_end=156
  _globals['_CQTORCHNEURONXSERVICE']._serialized_start=158
  _globals['_CQTORCHNEURONXSERVICE']._serialized_end=255
# @@protoc_insertion_point(module_scope)



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################
from torchvision.transforms.Transformer import Transformer
from proxies.mytorchvision.datasets.ImageFolderProxy import ImageFolderProxy
import torch.scaffolding.data_mgmt as data_mgmt
from torch.Tensor import Tensor
from torch.utils.data.Dataset import Dataset
from utils.MyTorchClient import MyTorchClient
from typing import Optional
import os

"""
The ImageFolder dataset class is a generic data loader where the images are arranged in this way:
root/dog/xxx.png
root/dog/xxy.png
root/dog/xxz.png
root/cat/123.png
root/cat/nsdf3.png
root/cat/asd932_.png

where 'dog', 'cat' are the class labels. The class labels are determined by the folder structure of the dataset.

The `datapath` can either be a path on the client or on the server. If it is on the client but not on the server,
the data will be uploaded to the server. If it is on the server, the data will be used from there (unless the 
OVERRIDE_SERVER_DATA flag is set).

If the datapath parameter is absolute, it will be assumed to be pointing to a directory on the client, as all server paths 
are relative to the server's data directory (which is set in the server code). In this case, the final folder in the full path will be
uploaded to the server. For example, if the datapath is /home/user/data/resnet, only the 'resnet' folder will be uploaded
to the server's data directory, where the folder `resnet` and any subdirectories will be created. However, if the folder
`resnet` already exists on the server, the data will not be uploaded again.

...Thus, ImageFolder('/home/user/data/resnet', transform) 
    - if folder does not exist on client:
        - raise FileNotFoundError
    - if folder exists on client:
        - see if folder `resnet` already exists on the server
            - if not, upload the 'resnet' folder to the server's data directory.


If the datapath parameter is relative, we will to see if the folder exists on the server (relative to the server's data
directory), and on the client (relative to the directory that contains the running scriopt). 
If the contents already exist on the server (and the OVERRIDE_SERVER_DATA flag is not set), the data will not be uploaded again.
If the contents do not exist on the server, the data will be uploaded to the server. Only the final folder in the relative path
will be uploaded to the server's data directory. For example, if the datapath is '../data/resnet', only the 'resnet' folder will be
uploaded to the server's data directory, where the folder `resnet` and any subdirectories will be created.

...Thus, ImageFolder('../data/resnet', transform)
    - if folder `resnet` already exists on the server
        - if yes, no data will be uploaded
    - else if folder exists on client:
        - upload the 'resnet' folder to the server's data directory.
    - if folder does not exist on client:
        - raise FileNotFoundError

"""

class ImageFolder(Dataset):
    def __init__(self, datapath: str, transform: Optional[Transformer] = None):
        super().__init__()
        self.datapath = datapath
        self.trailing_folder = os.path.basename(self.datapath)
        self.transform = transform
        self.proxy = ImageFolderProxy()
        # if no transform is provided, set transform_uuid to an empty string
        self.transform_uuid = self.transform.uuid if self.transform is not None else ""
        trailing_folder = os.path.basename(self.datapath) # e.g. 'resnet' if datapath is '/home/user/data/resnet'
        self.uuid, self.tuple_list, self.classes = self.createImageFolderOnServer()
        self.imgs = [img for img, label in self.tuple_list]
        self.targets = [label for img, label in self.tuple_list]

    def createImageFolderOnServer(self):
        if self.must_upload_data():
            # see if folder exists on the client
            if not os.path.exists(self.datapath):
                raise FileNotFoundError(f"Path `{self.datapath}` does not exist on the client")
            data_mgmt.upload_folder_to_server(self.datapath)

        image_folder_uuid, image_uuids, labels, classes = self.proxy.createImageFolderOnServer(self.trailing_folder, self.transform_uuid)
        # we want to generate a tuple list, where the tuple is (mytorch.Tensor, label)
        tuple_list =  [(Tensor(image_uuid), label) for image_uuid, label in zip(image_uuids, labels)]
        self.logger.info(f"ImageFolder with data from `{self.trailing_folder}` created with {len(tuple_list)} images.")
        return image_folder_uuid, tuple_list, classes

    def must_upload_data(self):
        if MyTorchClient.instance().get_overwrite_server_data():
            self.logger.info("OVERRIDE_SERVER_DATA flag is set. Uploading data to server.")
            return True

        if os.path.isabs(self.datapath): # if the datapath is absolute
            # see if folder exists on the client
            if not os.path.exists(self.datapath):
                raise FileNotFoundError(f"Path `{self.datapath}` does not exist on the client")
            # check if the folder exists on the server
            if data_mgmt.folder_exists_on_server(self.trailing_folder):
               self.logger.info(f"Data folder {self.trailing_folder} already exists on the server, no data will be uploaded")
               return False
            return True

        else: # if the datapath is relative
            on_server = data_mgmt.folder_exists_on_server(self.trailing_folder)
            if on_server:
                self.logger.info(f"Data folder `{self.trailing_folder}` already exists on the server, no data will be uploaded")
                return False
            # see if folder exists on the client
            if not os.path.exists(self.datapath):
                raise FileNotFoundError(f"Path `{self.datapath}` does not exist on the client")
            return True

    # for indexing, e.g. image_folder[0]
    def __getitem__(self, idx):
        # return index of self.tuple_list
        return self.tuple_list[idx]

    def __array__(self):
        return self.tuple_list

    # support iteration
    def __iter__(self):
        return iter(self.tuple_list)

    def __len__(self):
        # Return the length of the internal array
        return len(self.tuple_list)



# Import statements in transformers/__init__.py
from .models.auto.AutoModelForCausalLM import AutoModelForCausalLM
from .models.auto.AutoModelForSequenceClassification import AutoModelForSequenceClassification
from .models.auto.AutoModelForSeq2SeqLM import AutoModelForSeq2SeqLM
from .AutoTokenizer import AutoTokenizer
from .PreTrainedTokenizerFast import PreTrainedTokenizerFast
from .PreTrainedTokenizerBase import PreTrainedTokenizerBase
from .BertTokenzier import BertTokenizer
from .RobertaTokenizer import RobertaTokenizer
from .BertForSequenceClassification import BertForSequenceClassification
from .RobertaForSequenceClassification import RobertaForSequenceClassification
from .T5Tokenizer import T5Tokenizer
from .T5ForConditionalGeneration import T5ForConditionalGeneration
from .XLNetTokenizer import XLNetTokenizer
from .XLNetForSequenceClassification import XLNetForSequenceClassification



###############################################################################
# File: mytorch/client/torch/nn/LSTM.py
###############################################################################
from torch.nn.Module import Module
from proxies.mytorch.nn.lstm_proxy import LSTMProxy
from torch.Tensor import Tensor

class LSTM(Module):
    """
    An LSTM Module that registers a server-side LSTM object and delegates
    forward calls to it. We can call forward(input, hx) multiple times
    with persistent server-side weights.
    """

    def __init__(self, input_size, hidden_size, num_layers=1, bias=True, batch_first=False,
                 dropout=0.0, bidirectional=False):
        # We call super().__init__() so a new server-side “Module” UUID can be tracked.
        super().__init__(uuid=None)  
        
        # Create our specialized LSTM proxy
        self.proxy = LSTMProxy()

        # Instruct the server to create a brand new torch.nn.LSTM(...) object
        new_uuid = self.proxy.create_lstm_on_server(
            input_size, hidden_size, num_layers, bias, batch_first, dropout, bidirectional
        )

        # Our parent Module class expects us to set self.uuid to the server’s object
        self.set_uuid(new_uuid)

    def forward(self, input: Tensor, hx=None):
        """
        In real PyTorch, LSTM returns (output, (h_n, c_n)).
        We'll do the same by delegating to LSTMProxy.forward(...).
        """
        return self.proxy.forward(self.uuid, input, hx)




import json
from pprint import pprint

TOTAL_DURATION = 'total_duration'
CALL_COUNT = 'call_count'
NETWORK_DURATION = 'network_duration'
SERVER_DURATION = 'server_duration'

class TimingData:
    def __init__(self):
        self._run_data = {}  # key = run_id
        self._start_times = {}  # key = run_id
        self._end_times = {}  # key = run_id

    def start_time(self, time, run_id):
        self._start_times[run_id] = time

    def end_time(self, time, run_id):
        self._end_times[run_id] = time

    def dump_data(self, title, run_id):
        print(f"______________________________________ {title}")
        pprint(self._run_data[run_id])
        print("______________________________________")

    def init_from_json(self, json_string, run_id):
        timing_data = json.loads(json_string.value)
        self._run_data[run_id] = timing_data

    def add_method_duration(self, run_id, method_name, duration):
        if run_id not in self._run_data:
            self._run_data[run_id] = {}

        if method_name not in self._run_data[run_id]:
            self._run_data[run_id][method_name] = {
                TOTAL_DURATION: duration,
                CALL_COUNT: 1,
                NETWORK_DURATION: 0,  # not yet calculated
                SERVER_DURATION: 0  # not yet calculated
            }
            return

        method_stats = self._run_data[run_id][method_name]
        method_stats[CALL_COUNT] += 1
        method_stats[TOTAL_DURATION] += duration

    # ONLY used to present data to humans,
    def transform_and_round_ms(self, value):
        return round(value * 1000, 2)

    def transform_and_round_sec(self, value):
        return round(value, 2)

    def clear_statistics(self, run_id=None):
        if run_id:
            self._run_data.pop(run_id, None)
            self._start_times.pop(run_id, None)
            self._end_times.pop(run_id, None)
        else:
            self._run_data.clear()
            self._start_times.clear()
            self._end_times.clear()

    def incorporate_end_of_run_server_data(self, server_data: 'TimingData', run_id):
        if run_id not in self._run_data:
            raise ValueError(f"Run ID '{run_id}' not found in client data.")
        if run_id not in server_data._run_data:
            raise ValueError(f"Run ID '{run_id}' not found in server data.")

        total_server = 0
        client_methods = self._run_data[run_id]
        server_methods = server_data._run_data[run_id]

        #Fileter out methods with "timing_statistics" in the name, was causing problems
        client_method_set = {method for method in client_methods.keys() if "timing_statistics" not in method}
        server_method_set = {method for method in server_methods.keys() if "timing_statistics" not in method}

        if client_method_set != server_method_set:
            missing_in_client = server_method_set - client_method_set
            missing_in_server = client_method_set - server_method_set
            error_msg = "Mismatch in method names between client and server data."
            if missing_in_client:
                error_msg += f" Missing in client: {missing_in_client}."
            if missing_in_server:
                error_msg += f" Missing in server: {missing_in_server}."
            raise ValueError(error_msg)

        filtered_client_methods = {method: stats for method, stats in client_methods.items() if "timing_statistics" not in method}

        for method_name, client_stats in filtered_client_methods.items():
            server_stats = server_methods[method_name]
            server_total_duration = server_stats[TOTAL_DURATION]
            client_total_duration = client_stats[TOTAL_DURATION]
            client_stats[SERVER_DURATION] = server_total_duration
            network_duration = max(client_total_duration - server_total_duration, 0)
            client_stats[NETWORK_DURATION] = network_duration
            total_server += server_total_duration

    def get_statistics(self, run_id=None):
        if run_id:
            return self._get_statistics_for_run(run_id)
        else:
            return self._get_statistics_for_all_runs()

    def _get_statistics_for_all_runs(self):
        all_stats = {}
        for run_id in self._run_data.keys():
            if run_id not in self._start_times or run_id not in self._end_times:
                continue  # Skip this run_id (the initial run "DEFAULT NAME")
                          # TO DO, ELiminate this run_id entirely, refactor MyTorchCLient
            all_stats[run_id] = self._get_statistics_for_run(run_id)
        return all_stats
    
    def _get_statistics_for_run(self, run_id):
        if run_id not in self._run_data:
            return {}
        stats = {}
        for method, data in self._run_data[run_id].items():
            network_duration: float = data.get(NETWORK_DURATION)
            if network_duration is not None:
                network_duration = network_duration
            else:
                network_duration = 0.0
        
            server_duration: float = data.get(SERVER_DURATION)
            if server_duration is not None:
                server_duration = server_duration
            else:
                server_duration = 0.0
        
            total_duration: float = data.get(TOTAL_DURATION, 0)  # Default to 0 if not present
            call_count: int = data.get(CALL_COUNT, 0)  # Default to 0 if not present
            average_duration: float = (total_duration / call_count) if call_count else 0

            stats[method] = {
                CALL_COUNT: call_count,
                'average_duration': average_duration,
                TOTAL_DURATION: total_duration,
                SERVER_DURATION: server_duration,
                NETWORK_DURATION: network_duration
            }
        return stats
    
    def print_statistics(self, summary_only):
        all_stats = self.get_statistics()
        for run_id, run_stats in all_stats.items():
            display_name = run_id.split("::")[0]
            total_calls = sum(stats[CALL_COUNT] for stats in run_stats.values())
            total_compute_time = sum(stats[SERVER_DURATION] for stats in run_stats.values() if isinstance(stats[SERVER_DURATION], (int, float)))
            total_network_time = sum(stats[NETWORK_DURATION] for stats in run_stats.values() if isinstance(stats[NETWORK_DURATION], (int, float)))
            total_time = total_compute_time + total_network_time

            compute_percentage = (total_compute_time / total_time * 100) if total_time > 0 else 0
            network_percentage = (total_network_time / total_time * 100) if total_time > 0 else 0

            # Ensure rounding is properly applied for display
            compute_percentage = round(compute_percentage,2)
            network_percentage = round(network_percentage,2)

            # Summary for each run
            print()
            print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
            print(f"Timing Data Summary for run {display_name}:")
            print()
            print(f"\tTotal Calls:\t\t{total_calls}")
            print(f"\tTotal Time:\t{self.transform_and_round_sec(total_time):7.2f} sec")
            print(f"\t...Network:\t{self.transform_and_round_sec(total_network_time):7.2f} sec")
            print(f"\t...Sever:\t{self.transform_and_round_sec(total_compute_time):7.2f} sec")
            print(f"\tPercentage in Compute: {compute_percentage}%")
            print(f"\tPercentage in Network: {network_percentage}%\n")

            if summary_only:
                continue

            # Method-wise details
            print()
            print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
            print()
            print(f"By Method: Time taken")
            print()
            for method_name, method_stats in run_stats.items():
                # if method name is timing_statistics, skip it
                if "timing_statistics" in method_name:
                    continue
                network_duration = self.transform_and_round_ms(method_stats[NETWORK_DURATION]) if isinstance(method_stats[NETWORK_DURATION], (int, float)) else "N/A"
                server_duration = self.transform_and_round_ms(method_stats[SERVER_DURATION]) if isinstance(method_stats[SERVER_DURATION], (int, float)) else "N/A"
                avg_duration = method_stats[TOTAL_DURATION] / method_stats[CALL_COUNT]
                percent_of_total_network = (method_stats[NETWORK_DURATION] / total_network_time * 100) if total_network_time > 0 else 0
                percent_of_total_compute = (method_stats[SERVER_DURATION] / total_compute_time * 100) if total_compute_time > 0 else 0

                print(f"\tMethod:\t{method_name}")
                print(f"\t\tCall Count:\t{method_stats[CALL_COUNT]}")
                print(f"\t\tAverage time:\t{self.transform_and_round_ms(avg_duration):7.2f} ms")
                print(f"\t\tTotal time:\t\t{self.transform_and_round_ms(method_stats[TOTAL_DURATION]):7.2f} ms")
                print(f"\t\t...Network:\t\t{network_duration:7.2f} ms" if isinstance(method_stats[NETWORK_DURATION], (int, float)) else "      Network Duration: N/A")
                print(f"\t\t...Server:\t\t{server_duration:7.2f} ms" if isinstance(method_stats[SERVER_DURATION], (int, float)) else "      Server Duration: N/A")
                print("\t\tPercentage of all network time:\t{0:7.2f}%".format(percent_of_total_network))
                print("\t\tPercentage of all server time:\t{0:7.2f}%".format(percent_of_total_compute))
                print()



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from utils.logger import Logger
from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl.mytorchvision.transforms import transforms_pb2, transforms_pb2_grpc
from gRPC_impl import shared_msg_types_pb2

class ToTensorProxy:
    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = transforms_pb2_grpc.TransformsServiceStub(self.channel)
        self.logger = Logger.get_logger()

    @wrap_with_error_handler
    def createTransformOnServer(self):
        request = shared_msg_types_pb2.Empty()
        response: transforms_pb2.GrpcTransform = self.stub.CreateToTensorTransform(request)
        return response.uuid


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

"""
This script creates a subset of images from a larger dataset directory.
The script randomly selects a subset of images and copies them to a new directory while preserving the original directory structure.
The new subset can be used when creating an ImageFolder where the larger dataset is too unwieldy.
"""

import os
import shutil
import numpy as np

# Define the data paths
DATAPATH = "../../test_data/resnet_infer"
NEW_DATAPATH = "../../test_data/resnet_infer_416"
SUBSET_SIZE = 416  # Make this divisible by BATCH_SIZE (e.g. 32)

# Ensure the new data directory exists
if not os.path.exists(NEW_DATAPATH):
    os.makedirs(NEW_DATAPATH)

# Print the absolute path to verify it's correct
print(f"Absolute DATAPATH: {os.path.abspath(DATAPATH)}")

# List all image files in the dataset directory, keeping category subdirectories
all_files = []
for root, dirs, files in os.walk(DATAPATH):
    abs_root = os.path.abspath(root)
    print(f"Walking through: {abs_root}")
    print(f"Directories: {dirs}")
    print(f"Files: {files}")
    for file in files:
        if file.endswith(('jpg', 'jpeg', 'png', 'JPEG', 'JPG', 'PNG')):  # Add other image extensions if necessary
            file_path = os.path.join(root, file)
            print(f"Found file: {file_path}")  # Debug print to show the file path
            all_files.append(file_path)

print(f"Total files found: {len(all_files)}")  # Debug print

# Check if all_files is empty
if not all_files:
    raise ValueError("No image files found in the specified DATAPATH.")

# Select a random subset of the files
subset_files = np.random.choice(all_files, size=SUBSET_SIZE, replace=False)

# Copy the selected files to the new directory, preserving subdirectory structure
for file_path in subset_files:
    # Determine the relative path to preserve the directory structure
    relative_path = os.path.relpath(file_path, DATAPATH)
    new_file_path = os.path.join(NEW_DATAPATH, relative_path)

    # Ensure the target directory exists
    new_file_dir = os.path.dirname(new_file_path)
    if not os.path.exists(new_file_dir):
        os.makedirs(new_file_dir)

    # Copy the file
    shutil.copy(file_path, new_file_path)

print(f"Copied {SUBSET_SIZE} files to {NEW_DATAPATH}")



###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from proxies.mytorch.nn.functional.functional_proxy import FunctionalProxy
from torch.Tensor import Tensor

def softmax(input: Tensor, dim: int) -> Tensor:
    proxy = FunctionalProxy()
    return proxy.softmax(input, dim)

#def relu(input: Tensor, inplace: bool = False) -> Tensor:
#    proxy = FunctionalProxy()
#    return proxy.relu(input, inplace)


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

'''
Because all of the HuggingFace tokenizers have the same API, we can create a single proxy class that can be used to interact with any of them.
'''

from connection_utils.server_connection import ServerConnection, wrap_with_error_handler
from gRPC_impl import shared_msg_types_pb2
from gRPC_impl.mytransformers import HuggingFaceTokenizer_pb2, HuggingFaceTokenizer_pb2_grpc
from utils.logger import Logger
from torch.Tensor import Tensor

class HuggingFaceTokenizerProxy:

    def __init__(self):
        self.channel = ServerConnection.get_active_connection()
        self.stub = HuggingFaceTokenizer_pb2_grpc.HuggingFaceTokenizerServiceStub(self.channel)
        self.logger = Logger.get_logger()

    @wrap_with_error_handler
    def getAutoTokenizer(self, pretrained_model_name_or_path, *inputs, **kwargs) -> tuple[str, dict]:
        request = HuggingFaceTokenizer_pb2.tokenizerFromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         inputs=inputs,
                                                         kwargs=kwargs)
        response = self.stub.getAutoTokenizer(request)
        special_tokens_dict = dict(response.special_tokens)
        return response.tokenizer_uuid, special_tokens_dict

    @wrap_with_error_handler
    def getBertTokenizer(self, pretrained_model_name_or_path, *inputs, **kwargs) -> tuple[str, dict]:
        request = HuggingFaceTokenizer_pb2.tokenizerFromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         inputs=inputs,
                                                         kwargs=kwargs)
        response = self.stub.getBertTokenizer(request)
        special_tokens_dict = dict(response.special_tokens)
        return response.tokenizer_uuid, special_tokens_dict

    @wrap_with_error_handler
    def getRobertaTokenizer(self, pretrained_model_name_or_path, *inputs, **kwargs) -> tuple[str, dict]:
        request = HuggingFaceTokenizer_pb2.tokenizerFromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         inputs=inputs,
                                                         kwargs=kwargs)
        response = self.stub.getRobertaTokenizer(request)
        special_tokens_dict = dict(response.special_tokens)
        return response.tokenizer_uuid, special_tokens_dict

    @wrap_with_error_handler
    def getT5Tokenizer(self, pretrained_model_name_or_path, *inputs, **kwargs) -> tuple[str, dict]:
        request = HuggingFaceTokenizer_pb2.tokenizerFromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         inputs=inputs,
                                                         kwargs=kwargs)
        response = self.stub.getT5Tokenizer(request)
        special_tokens_dict = dict(response.special_tokens)
        return response.tokenizer_uuid, special_tokens_dict

    @wrap_with_error_handler
    def getXLNetTokenizer(self, pretrained_model_name_or_path, *inputs, **kwargs) -> tuple[str, dict]:
        request = HuggingFaceTokenizer_pb2.tokenizerFromPretrained_request(pretrained_model_name_or_path=pretrained_model_name_or_path,
                                                         inputs=inputs,
                                                         kwargs=kwargs)
        response = self.stub.getXLNetTokenizer(request)
        special_tokens_dict = dict(response.special_tokens)
        return response.tokenizer_uuid, special_tokens_dict

    @wrap_with_error_handler
    def encode_plus(self, tokenizer_uuid, text, **kwargs):
        request = HuggingFaceTokenizer_pb2.encodePlus_request(tokenizer_uuid=tokenizer_uuid, text=text)
        for key, val in kwargs.items():
            # value can be a string, int, float, or bool
            if isinstance(val, str):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, string_value=val)])
            elif isinstance(val, bool):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, bool_value=val)])
            elif isinstance(val, int):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, int_value=val)])
            elif isinstance(val, float):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, float_value=val)])
        response = self.stub.encode_plus(request)
        # convert the response.data to a dictionary
        data = {}
        for pair in response.data:
            tensor = Tensor(uuid=pair.tensor_uuid)
            data[pair.key] = tensor
        return response.encoding_uuid, data

    @wrap_with_error_handler
    def decode(self, tokenizer_uuid, input_tensor_uuid, **kwargs):
        request = HuggingFaceTokenizer_pb2.decode_request(tokenizer_uuid=tokenizer_uuid, input_tensor_uuid=input_tensor_uuid)
        for key, val in kwargs.items():
            # value can be a string, int, float, or bool
            if isinstance(val, str):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, string_value=val)])
            elif isinstance(val, bool):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, bool_value=val)])
            elif isinstance(val, int):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, int_value=val)])
            elif isinstance(val, float):
                request.kwargs.extend([shared_msg_types_pb2.keyValuePair(key=key, float_value=val)])
        response = self.stub.decode(request)
        return response.text

    @wrap_with_error_handler
    def set_pad_token(self, tokenizer_uuid, value):
        request = HuggingFaceTokenizer_pb2.setPadToken_request(tokenizer_uuid=tokenizer_uuid, pad_token=value)
        self.stub.set_pad_token(request)
        return


###############################################################################
# Copyright (c) 2024 MyTorch Systems Inc. All rights reserved.
###############################################################################

from proxies.mytorch.cuda.cuda_proxy import CudaProxy
from collections import namedtuple

# The return type of the get_device_properties function is a named tuple
_CudaDeviceProperties = namedtuple('_CudaDeviceProperties', [
    'name', 'total_memory', 'multi_processor_count', 'available_memory', 'allocated_memory',
    'reserved_memory', 'max_memory_allocated', 'max_memory_reserved'
])

def is_available():
    return CudaProxy().is_available()

def empty_cache():
    CudaProxy().empty_cache()

def memory_allocated():
    return CudaProxy().memory_allocated()

def memory_reserved():
    return CudaProxy().memory_reserved()

def get_device_properties(device):
    device_info = CudaProxy().get_device_properties(device)

    # Filter the dictionary to only include keys that are in the namedtuple
    filtered_device_info = {k: v for k, v in device_info.items() if k in _CudaDeviceProperties._fields}

    # Instantiate the namedtuple using ** to unpack the dictionary
    device_properties = _CudaDeviceProperties(**filtered_device_info)
    return device_properties

def get_device_name():
    return CudaProxy().get_device_name()





# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mytorch/jit/jit.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gRPC_impl import shared_msg_types_pb2 as shared__msg__types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15mytorch/jit/jit.proto\x12\x07mytorch\x1a\x16shared_msg_types.proto\"=\n\x12ModelIDandTensorID\x12\x12\n\nmodel_uuid\x18\x01 \x01(\t\x12\x13\n\x0btensor_uuid\x18\x02 \x01(\t\"/\n\x0bSaveRequest\x12\x12\n\nmodel_uuid\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t2\x98\x01\n\nJitService\x12\x32\n\x05trace\x12\x1b.mytorch.ModelIDandTensorID\x1a\x0c.shared.UUID\x12+\n\x04save\x12\x14.mytorch.SaveRequest\x1a\r.shared.Empty\x12)\n\x04load\x12\x13.shared.StringValue\x1a\x0c.shared.UUIDb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mytorch.jit.jit_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_MODELIDANDTENSORID']._serialized_start=58
  _globals['_MODELIDANDTENSORID']._serialized_end=119
  _globals['_SAVEREQUEST']._serialized_start=121
  _globals['_SAVEREQUEST']._serialized_end=168
  _globals['_JITSERVICE']._serialized_start=171
  _globals['_JITSERVICE']._serialized_end=323
# @@protoc_insertion_point(module_scope)
